unit UMainForm;

interface

uses
  Winapi.Windows, Winapi.Messages, Winapi.Dwmapi, System.Types, System.SysUtils,
  System.Classes, FMX.Types3D, Vcl.Graphics, Vcl.Controls, Vcl.Dialogs, Vcl.Forms,
  LibOVR, LibOVR.GL, OpenGL.Apis, OpenGL.Context.Win;

type
  TForm1 = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormResize(Sender: TObject);
  private
    { Private 錾 }
    FHeadPos: TVector3D;
    FHmd: TOculusRift;
    FDC: HDC;
    FRC: HGLRC;
    FSampler: GLuint;
    FVertexShader: GLuint;
    FFragmentShader: GLuint;
    FProgram: GLuint;
    FMVPMatrixLocation: GLint;
    FPositionLocation: GLint;
    FColorLocation: GLint;
    FFrameBuffer: GLuint;
    FTexture: GLuint;
    FRenderBuffer: GLuint;
    FVertexArray: GLuint;
    FPositionBuffer: GLuint;
    FColorBuffer: GLuint;
    FWindowSize: TRect;
    FRotationAngle: Integer;
    procedure AppIdle(Sender: TObject; var Done: Boolean);
  public
    { Public 錾 }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
const
  vertexSource: PAnsiChar =
    '#version 330'#10 +
    'uniform mat4 MVPMatrix;'#10 +
    'in vec3 VertexPosition;'#10 +
    'in vec4 VertexColor;'#10 +
    'out vec4 Color;'#10 +
    'void main()'#10 +
    '{'#10 +
    '  gl_Position = MVPMatrix * vec4(VertexPosition, 1.0);'#10 +
    '  Color = VertexColor;'#10 +
    '}';
  fragmentSource: PAnsiChar =
    '#version 330'#10 +
    'in vec4 Color;'#10 +
    'out vec4 FragColor;'#10 +
    'void main()'#10 +
    '{'#10 +
    '  FragColor = Color;'#10+
    '}';
  vertices: array[0..8] of GLfloat = (
    0,1,0,
    -1,-1,0,
    1,-1,0
  );
  colors: array[0..11] of GLfloat = (
    1,1,0,1,
    0,1,1,1,
    1,0,1,1
  );
var
  status: GLint;
  renderTargetSize: ovrSizei;
begin
  if not TOculusRift.Available then
  begin
    ShowMessage('OculusRift not found');
    Application.Terminate;
    Exit;
  end;

  // Initialize OpenGL

  FDC := GetDC(Handle);
  FRC := CreateWGLRenderingContext(FDC, TGLOption.Create(8,8,8,8,24,8,True), 4, 0).RenderingContext;
  wglMakeCurrent(FDC, FRC);

  glGenSamplers(1, @FSampler);
  glBindSampler(0, FSampler);
  glSamplerParameteri(FSampler, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glSamplerParameteri(FSampler, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glSamplerParameteri(FSampler, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glSamplerParameteri(FSampler, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  FVertexShader := glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(FVertexShader, 1, @vertexSource, nil);
  glCompileShader(FVertexShader);
  glGetShaderiv(FVertexShader, GL_COMPILE_STATUS, @status);
  if status = GL_FALSE then
    raise Exception.Create('Vertex shader compiling error');

  FFragmentShader := glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(FFragmentShader, 1, @fragmentSource, nil);
  glCompileShader(FFragmentShader);
  glGetShaderiv(FFragmentShader, GL_COMPILE_STATUS, @status);
  if status = GL_FALSE then
    raise Exception.Create('Fragment shader compiling error');

  FProgram := glCreateProgram;
  glAttachShader(FProgram, FVertexShader);
  glAttachShader(FProgram, FFragmentShader);
  glLinkProgram(FProgram);
  glGetProgramiv(FProgram, GL_LINK_STATUS, @status);
  if status = GL_FALSE then
    raise Exception.Create('Program link error');

  glUseProgram(FProgram);
  FMVPMatrixLocation := glGetUniformLocation(FProgram, 'MVPMatrix');
  FPositionLocation := glGetAttribLocation(FProgram, 'VertexPosition');
  FColorLocation := glGetAttribLocation(FProgram, 'VertexColor');

  glGenVertexArrays(1, @FVertexArray);
  glBindVertexArray(FVertexArray);

  glGenBuffers(1, @FPositionBuffer);
  glBindBuffer(GL_ARRAY_BUFFER, FPositionBuffer);
  glBufferData(GL_ARRAY_BUFFER, SizeOf(vertices), @vertices, GL_STATIC_DRAW);
  glVertexAttribPointer(FPositionLocation, 3, GL_FLOAT, GL_FALSE, 0, nil);
  glEnableVertexAttribArray(FPositionLocation);

  glGenBuffers(1, @FColorBuffer);
  glBindBuffer(GL_ARRAY_BUFFER, FColorBuffer);
  glBufferData(GL_ARRAY_BUFFER, SizeOf(colors), @colors, GL_STATIC_DRAW);
  glVertexAttribPointer(FColorLocation, 4, GL_FLOAT, GL_FALSE, 0, nil);
  glEnableVertexAttribArray(FColorLocation);

  glBindVertexArray(0);

  // Initialize OculusRift

  FHmd := TOculusRift.Create(Handle);

  renderTargetSize := FHmd.GetRenderTargetSize;
  glGenFramebuffers(1, @FFrameBuffer);
  glBindFramebuffer(GL_FRAMEBUFFER, FFrameBuffer);
  glGenTextures(1, @FTexture);
  glBindTexture(GL_TEXTURE_2D, FTexture);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, renderTargetSize.w, renderTargetSize.h, 0, GL_RGBA, GL_UNSIGNED_BYTE, nil);
  glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, FTexture, 0);
  glGenRenderbuffers(1, @FRenderBuffer);
  glBindRenderbuffer(GL_RENDERBUFFER, FRenderBuffer);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, renderTargetSize.w, renderTargetSize.h);
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, FRenderBuffer);
  glBindFramebuffer(GL_FRAMEBUFFER, 0);
  glClearColor(0, 0, 1, 1);
  glClearDepth(1.0);
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);

  FHmd.InitializeOpenGL(FTexture);
  FHmd.Resize(ClientWidth, ClientHeight);

  FHeadPos := Vector3D(0, 0, 1);

  Application.OnIdle := AppIdle;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  if FHmd = nil then Exit;
  glDeleteVertexArrays(1, @FVertexArray);
  glDeleteBuffers(1, @FPositionBuffer);
  glDeleteFramebuffers(1, @FFrameBuffer);
  glDeleteRenderbuffers(1, @FRenderBuffer);
  glDeleteTextures(1, @FTexture);
  glDeleteShader(FVertexShader);
  glDeleteShader(FFragmentShader);
  glDeleteProgram(FProgram);
  glDeleteSamplers(1, @FSampler);
  wglMakeCurrent(0, 0);
  wglDeleteContext(FRC);
  ReleaseDC(Handle, FDC);
  FHmd.Free;
end;

procedure TForm1.AppIdle(Sender: TObject; var Done: Boolean);
begin
  Done := False;
  FHmd.BeginFrame;
  try
    glBindFrameBuffer(GL_FRAMEBUFFER, FFrameBuffer);
    glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

    FHmd.RenderFrame(
      procedure(const APose: ovrPosef; const AFov: ovrFovPort;
        const AViewAdjust: ovrVector3f; const AViewport: ovrRecti)
      var
        q: TQuaternion3D;
        finalRollPitchYaw: TMatrix3D;
        finalUp, finalForward, shiftedEyePos: TVector3D;
        model, view, proj, mvp: TMatrix3D;
      begin
        q.ImagPart := Vector3D(APose.Orientation.x, APose.Orientation.y, APose.Orientation.z);
        q.RealPart := APose.Orientation.w;
        finalRollPitchYaw := QuaternionToMatrix(q);
        finalUp := Vector3D(0,1,0,0) * finalRollPitchYaw;
        finalForward := Vector3D(0,0,-1,0) * finalRollPitchYaw;
        shiftedEyePos := FHeadPos + Vector3D(TPoint3D(APose.Position)) + Vector3D(TPoint3D(AViewAdjust)) * finalRollPitchYaw;
        model := CreateRotationMatrix3D(Vector3D(0,1,0), FRotationAngle/100);
        view := MatrixLookAtRH(shiftedEyePos, shiftedEyePos + finalForward, finalUp);
        proj := TMatrix3D(ovrMatrix4f_Projection(AFov, 0.01, 10000.0, True));
        TransposeMatrix3D(proj);
        mvp := model * view * proj;

        glViewport(AViewport.Pos.x, AViewport.Pos.y, AViewport.Size.w, AViewport.Size.h);
        glUniformMatrix4fv(FMVPMatrixLocation, 1, GL_FALSE, @mvp);
        glBindVertexArray(FVertexArray);
        glEnableVertexAttribArray(FColorLocation);
        glDrawArrays(GL_TRIANGLES, 0, 3);
      end);
    glBindFrameBuffer(GL_FRAMEBUFFER, 0);
    glBindVertexArray(0);
    FRotationAngle := (FRotationAngle + 1) mod 36000;
  finally
    FHmd.EndFrame;
  end;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE: Application.Terminate;
    VK_RETURN:
    begin
      if BorderStyle = bsNone then
      begin
        DwmEnableComposition(DWM_EC_ENABLECOMPOSITION);
        BorderStyle := bsSizeable;
        BoundsRect := FWindowSize;
      end
      else begin
        DwmEnableComposition(DWM_EC_DISABLECOMPOSITION);
        BorderStyle := bsNone;
        FWindowSize := BoundsRect;
        BoundsRect := Monitor.BoundsRect;
      end;
    end;
    Ord('R'): FHmd.RecenterPose;
  end;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  if FHmd = nil then Exit;
  FHmd.Resize(ClientWidth, ClientHeight);
end;

end.
